var {
  THREE,
  Edge,
  Node,
  OrbitControls,
  d3,
  ee,
  DEF_NODE_COLOR,
  DEF_NODE_HIGHLIGHT,
  DEF_ADJACENT_HIGHLIGHT,
  DEF_NODE_HTEXTURE,
  DEF_NODE_TEXTURE,
  DEF_NODE_NO_HTEX,
  DEF_NODE_NO_TEX,
  DEF_NODE_NO_HCOLOR,
  DEF_NODE_NO_COLOR,
  DEF_NODE_NO_HIGHLIGHT,
  DEF_NODE_SHAPE,
  DEF_NODE_HSHAPE,
  DEF_NODE_SIZE,
  DEF_LAYOUT,
  DEF_LINES,
  DEF_MAP,
  DEF_TEXT_SCALE,
  DEF_TEXT_SIZE,
  DEF_TEXT_X_OFFSET,
  DEF_TEXT_Y_OFFSET,
  DEF_TEXT_BACK_COLOR,
  DEF_TEXT_OPACITY,
  TEXT_WIREFRAME
} = require('./imports').default;
var Utils = require('./utils/utils');

module.exports = function (self) {
  /**
  * Creates or converts ngraph or d3 graph
  */
  self.setupGraph = function () {
    var createNGraph = require('ngraph.graph');
    self.graph = createNGraph();
    if (self.inGraph) {
      self.updateGraph(self.inGraph);
    }
  }

  /**
   *  Create layout using d3
   */
  self.setupLayout = function () {
    if (self.options.layout == "ngraph") {
      self.force = require('ngraph.forcelayout')(self.graph);
    } else if (self.options.layout == "d3") {
      self.force = d3.forceSimulation()
        .force("charge", d3.forceManyBody().strength(-1))
        .force("link", d3.forceLink())
        .force("center", d3.forceCenter())
        .stop();

        // set alpha decay to be lower than the default 0.0228
        // so force layout does not prematurely stop
        self.force.alphaDecay(0.005)

        // set velocity decay to be lower than the default 0.4
        // so nodes move to their final locations in fewer ticks
        self.force.velocityDecay(0.1)
    }
  };

  /**
   *  Create initial camera and parameters
   */
  self.setupCamera = function () {
    self.ccamera = new THREE.PerspectiveCamera(self.fov, self.width / self.height, self.near, self.far);
    self.ccamera.position.z = 200;
    self.ccamera2 = new THREE.PerspectiveCamera(self.fov, self.width / self.height, self.near, self.far);
    self.ccamera2.position.z = self.renderWidth * 2;
    var controls = new OrbitControls(self.ccamera, self.element);
    controls.minDistance = 10;
    controls.maxDistance = self.renderWidth * 2;
    controls.target.set(0, 0, 0);
    controls.enableRotate = false;
    self.controls = controls;
  };

  /**
   *  Create initial scene geometry and attributes
   */
  self.setupGeometry = function () {
    self.scene = new THREE.Scene();

    self.points = new THREE.BufferGeometry();
    self.nodes = new THREE.Group();
    self.scene.add(self.nodes);

    if (DEF_LINES == "fancy") {
      self.edges = [];
      self.edgeCount = 0;
    } else {
      self.edges = new THREE.BufferGeometry();
      var material = new THREE.LineBasicMaterial({
        linewidth: 2,
        color: 0xffffff,
        vertexColors: THREE.VertexColors,
        shading: THREE.FlatShading
      });
      self.line = new THREE.LineSegments(self.edges, material);
      self.line.frustumCulled = false;
      self.scene.add(self.line);

      var positions = new THREE.BufferAttribute(new Float32Array(self.MAX_LINES * 3), 3);
      var colors = new THREE.BufferAttribute(new Float32Array(self.MAX_LINES * 3), 3);

      self.edges.addAttribute('position', positions);
      self.edges.addAttribute('color', colors);
      self.drawCount = 0;
    }

    var rectShape = new THREE.Shape();
    rectShape.moveTo(0, 0);
    rectShape.lineTo(0, 1);
    rectShape.lineTo(1, 1);
    rectShape.lineTo(1, 0);
    rectShape.lineTo(0, 0);
    rectShape.lineTo(0, 1);
    var rect = new THREE.ShapeGeometry(rectShape);
    self.selectBox = new THREE.Line(rect, new THREE.LineBasicMaterial({linewidth: 3, color: 0x3399AA}));
    self.selectBox.visible = false;
    self.scene.add(self.selectBox);

    self.boundaries = new THREE.Line(rect, new THREE.LineBasicMaterial({linewidth: 3, color: 0x3399AA}));
    self.scene.add(self.boundaries);
    self.setBoundarySize(self.renderWidth * 2);
  };

  /**
   *  Creates listeners and events for selecting nodes
   */
  self.setupSelect = function () {
    self.points.addAttribute('position', new THREE.BufferAttribute(new Float32Array(20 * 3), 3));
    self.points.computeBoundingSphere();
    self.nodeCount = 0;
    var mouseHandler = function (callback) {
      return function (event) {
        event.preventDefault();
        var coords = Utils.relMouseCoords(event, this);
        var mouseX = (coords.x / self.width) * 2 - 1;
        var mouseY = 1 - (coords.y / self.height) * 2;
        var mousePosition = new THREE.Vector3(mouseX, mouseY, 1);
        mousePosition.unproject(self.ccamera);
        var dir = mousePosition.sub(self.ccamera.position).clone();
        mousePosition.normalize();

        // Determine intersects
        var raycaster = new THREE.Raycaster(self.ccamera.position, mousePosition);

        // Find real location of mouse
        var distance = -self.ccamera.position.z / dir.z;
        var pos = self.ccamera.position.clone().add(dir.multiplyScalar(distance));

        var intersects = raycaster.intersectObjects(self.nodes.children);
        if (intersects.length) {
          var nodeIndex = intersects[0].object.index;
          self.oldIntersect = nodeIndex;
          callback(self.graph.getNode(nodeIndex), pos.x, pos.y, event.button, event.ctrlKey);
        } else {
          callback(null, pos.x, pos.y, event.button, event.ctrlKey);
        }
      };
    };

    //Add listeners to web page
    self.element.addEventListener(
      'mousemove', mouseHandler(self.onMouseMove), false);
    self.element.addEventListener(
      'mousedown', mouseHandler(self.onMouseDown), false);
    self.element.addEventListener(
      'mouseup', mouseHandler(self.onMouseUp), false);
    self.element.addEventListener(
      'contextmenu', mouseHandler(self.onRightClick), false);
  };
}

